@@ -1,10 +1,12 @@
+# -*- coding: utf-8 -*-
+from __future__ import unicode_literals
+from __future__ import print_function
 """
 This module implements a friendly (well, friendlier) interface between the raw JSON
 responses from JIRA and the Resource/dict abstractions provided by this library. Users
 will construct a JIRA object as described below. Full API documentation can be found
 at: https://jira-python.readthedocs.org/en/latest/
 """
-from __future__ import print_function
 
 from functools import wraps
 
@@ -16,6 +18,7 @@ import os
 import re
 import sys
 import string
+import warnings
 import tempfile
 import logging
 import requests
@@ -23,7 +26,7 @@ import json
 import warnings
 
 from six import string_types
-from six.moves.html_parser import HTMLParser
+from six.moves import html_parser
 from six import print_ as print
 
 from requests_oauthlib import OAuth1
@@ -50,6 +53,10 @@ if 'pydevd' not in sys.modules:
     except NotImplementedError:
         pass
 
+encoding = sys.getdefaultencoding()
+if encoding != 'UTF8':
+    warnings.warn("Python default encoding is '%s' instead of 'UTF8' which means that there is a big change of having problems. Possible workaround http://stackoverflow.com/a/17628350/99834" % encoding)
+
 
 def translate_resource_args(func):
     """
@@ -97,6 +104,7 @@ class JIRA(object):
         "rest_api_version": "2",
         "verify": True,
         "resilient": False,
+        "async": False,
         "headers": {
             'X-Atlassian-Token': 'nocheck',
             #'Cache-Control': 'no-cache',
@@ -107,7 +115,7 @@ class JIRA(object):
 
     JIRA_BASE_URL = '{server}/rest/api/{rest_api_version}/{path}'
 
-    def __init__(self, options=None, basic_auth=None, oauth=None, validate=None):
+    def __init__(self, server=None, options=None, basic_auth=None, oauth=None, validate=None, async=False):
         """
         Construct a JIRA client instance.
 
@@ -121,6 +129,8 @@ class JIRA(object):
 
         For quick command line access to a server, see the ``jirashell`` script included with this distribution.
 
+        The easiest way to instantiate is using j = JIRA("https://jira.atlasian.com")
+
         :param options: Specify the server and properties this client will use. Use a dict with any
             of the following properties:
             * server -- the server address and context path to use. Defaults to ``http://localhost:2990/jira``.
@@ -138,9 +148,15 @@ class JIRA(object):
             JIRA in the OAuth application link)
         :param validate: If true it will validate your credentials first. Remember that if you are accesing JIRA
             as anononymous it will fail to instanciate.
+        :param async: To enable async requests for those actions where we implemented it, like issue update() or delete().
+        Obviously this means that you cannot rely on the return code when this is enabled.
         """
         if options is None:
             options = {}
+        if server:
+            options['server'] = server
+        if async:
+            options['async'] = async
 
         self._options = copy.copy(JIRA.DEFAULT_OPTIONS)
 
@@ -172,6 +188,7 @@ class JIRA(object):
 
     def __del__(self):
         # that is performing the logout by killing the session
+        self.async_do()
         self.kill_session()
 
 # Information about this client
@@ -214,6 +231,7 @@ class JIRA(object):
         :return:
         """
         if hasattr(self._session, '_async_jobs'):
+            logging.info("Executing async %s jobs found in queue by using %s threads..." % (len(self._session._async_jobs), size))
             grequests.map(self._session._async_jobs, size=size)
 
 # Application properties
@@ -262,7 +280,7 @@ class JIRA(object):
 
         headers = copy.deepcopy(self._options['headers'])
         headers['content-type'] = 'application/json;charset=UTF-8'
-        headers['accept']='application/json;charset=UTF-8'
+        headers['accept'] = 'application/json;charset=UTF-8'
         r = self._session.get(url, headers=headers)
 
         raise_on_error(r)
@@ -518,7 +536,7 @@ class JIRA(object):
         :param fields: comma-separated string of issue fields to include in the results
         :param expand: extra information to fetch inside each resource
         """
-        issue = Issue(self._options, self._session)
+        issue = Issue(self._options, self._session, self)
 
         params = {}
         if fields is not None:
@@ -598,7 +616,7 @@ class JIRA(object):
     @translate_resource_args
     def assign_issue(self, issue, assignee):
         """
-        Assign an issue to a user.
+        Assign an issue to a user. None will set it to unassigned. -1 will set it to Automatic.
 
         :param issue: the issue to assign
         :param assignee: the user to assign the issue to
@@ -708,9 +726,9 @@ class JIRA(object):
         if type(destination) == Issue:
 
             data['object'] = {
-                    'title': str(destination),
-                    'url': destination.permalink()
-                }
+                'title': str(destination),
+                'url': destination.permalink()
+            }
 
             for x in self.applicationlinks():
                 if x['application']['displayUrl'] == destination._options['server']:
@@ -725,12 +743,19 @@ class JIRA(object):
                 data['globalId'] = globalId
             if application is not None:
                 data['application'] = application
+            data['object'] = destination
 
         if relationship is not None:
             data['relationship'] = relationship
 
+        for x in self.applicationlinks():
+            if x['displayUrl'] == self._options['server']:
+                data['globalId'] = "appId=%s&issueId=%s" % (x['id'], destination.raw['id'])
+                data['application'] = {'name': x['name'], 'type': "com.atlassian.jira"}
+                break
+
         url = self._get_url('issue/' + str(issue) + '/remotelink')
-        r = self._session.post(url, headers={'content-type': 'application/json', 'accept': 'application/json'}, data=json.dumps(data))
+        r = self._session.post(url, headers={'content-type': 'application/json'}, data=json.dumps(data))
         raise_on_error(r)
 
         remote_link = RemoteLink(self._options, self._session, raw=json.loads(r.text))
@@ -848,7 +873,8 @@ class JIRA(object):
         """
         url = self._get_url('issue/' + str(issue) + '/watchers')
         params = {'username': watcher}
-        self._session.delete(url, params=params)
+        result = self._session.delete(url, params=params)
+        return result
 
     @translate_resource_args
     def worklogs(self, issue):
@@ -873,7 +899,7 @@ class JIRA(object):
 
     @translate_resource_args
     def add_worklog(self, issue, timeSpent=None, adjustEstimate=None,
-                    newEstimate=None, reduceBy=None, comment=None):
+                    newEstimate=None, reduceBy=None, comment=None, started=None, user=None):
         """
         Add a new worklog entry on an issue and return a Resource for it.
 
@@ -883,6 +909,7 @@ class JIRA(object):
         time estimate of the issue. The value can either be ``new``, ``leave``, ``manual`` or ``auto`` (default).
         :param newEstimate: the new value for the remaining estimate field. e.g. "2d"
         :param reduceBy: the amount to reduce the remaining estimate by e.g. "2d"
+        :param started: Moment when the work is logged, if not specified will default to now
         :param comment: optional worklog comment
         """
         params = {}
@@ -898,7 +925,21 @@ class JIRA(object):
             data['timeSpent'] = timeSpent
         if comment is not None:
             data['comment'] = comment
+        elif user:
+            # we log user inside comment as it doesn't always work
+            data['comment'] = user
 
+        if started is not None:
+            # based on REST Browser it needs: "2014-06-03T08:21:01.273+0000"
+            data['started'] = started.strftime("%Y-%m-%dT%H:%M:%S.000%z")
+        if user is not None:
+            data['author'] = { "name": user,
+                               'self': self.JIRA_BASE_URL + '/rest/api/2/user?username=' + user,
+                               'displayName': user,
+                               'active': False
+                               }
+            data['updateAuthor'] = data['author']
+        # TODO: report bug to Atlassian: author and updateAuthor parameters are ignored.
         url = self._get_url('issue/{}/worklog'.format(issue))
         r = self._session.post(url, params=params, headers={'content-type': 'application/json'}, data=json.dumps(data))
         raise_on_error(r)
@@ -1718,7 +1759,8 @@ class JIRA(object):
             'id': '',
             'Preview': 'Preview',
         }
-        r = self._session.post(url, headers={'X-Atlassian-Token': 'nocheck', 'Cache-Control': 'no-cache, no-store, no-transform'}, data=payload)
+
+        r = self._session.post(url, headers=self._options['headers'], data=payload)
         open("/tmp/jira_email_user_%s.html" % user, "w").write(r.content)
         # return False
 
@@ -2025,9 +2067,9 @@ class GreenHopper(JIRA):
 
     GREENHOPPER_BASE_URL = '{server}/rest/greenhopper/1.0/{path}'
 
-    def __init__(self, options=None, basic_auth=None, oauth=None):
+    def __init__(self, options=None, basic_auth=None, oauth=None, async=None):
         self._rank = None
-        JIRA.__init__(self, options, basic_auth, oauth)
+        JIRA.__init__(self, options=options, basic_auth=basic_auth, oauth=oauth, async=None)
 
     def find(self, resource_format, ids=None):
         """
@@ -2068,21 +2110,63 @@ class GreenHopper(JIRA):
         return boards
 
     @translate_resource_args
-    def sprints(self, id):
+    def sprints(self, id, extended=False):
         """
         Get a list of sprint GreenHopperResources.
 
         :param id: the board to get sprints from
+        :param extended: fetch additional information like startDate, endDate, completeDate,
+        much slower because it requires an additional requests for each sprint
+
+        "id": 893,
+        "name": "iteration.5",
+        "state": "FUTURE",
+        "linkedPagesCount": 0,
+        "startDate": "None",
+        "endDate": "None",
+        "completeDate": "None",
+        "remoteLinks": []
+
         """
-        r_json = {}
-        try:
-            r_json = self._get_json('sprintquery/%s' % id, base=self.GREENHOPPER_BASE_URL)
-        except:
-            r_json = self._get_json('sprints/%s' % id, base=self.GREENHOPPER_BASE_URL)
+        r_json = self._get_json('sprintquery/%s?includeHistoricSprints=true&includeFutureSprints=true' % id, base=self.GREENHOPPER_BASE_URL)
+
+        if extended:
+            sprints = []
+            for raw_sprints_json in r_json['sprints']:
+                r_json = self._get_json('sprint/%s/edit/model' % raw_sprints_json['id'], base=self.GREENHOPPER_BASE_URL)
+                sprints.append(Sprint(self._options, self._session, r_json['sprint']))
+        else:
+            sprints = [Sprint(self._options, self._session, raw_sprints_json) for raw_sprints_json in r_json['sprints']]
 
-        sprints = [Sprint(self._options, self._session, raw_sprints_json) for raw_sprints_json in r_json['sprints']]
         return sprints
 
+    def sprints_by_name(self, id, extended=False):
+        sprints = {}
+        for s in self.sprints(id, extended=extended):
+            print(s.raw)
+            if s.name not in sprints:
+                sprints[s.name] = s.raw
+            else:
+                raise(Exception("Fatal error, duplicate Sprint Name (%s) found on board %s." % (s.name, id)))
+        return sprints
+
+    def update_sprint(self, id, name=None, startDate=None, endDate=None):
+        payload = {}
+        if name:
+            payload['name'] = name
+        if startDate:
+            payload['startDate'] = startDate
+        if endDate:
+            payload['startDate'] = endDate
+        # if state:
+        #    payload['state']=state
+
+        url = self._get_url('sprint/%s' % id, base=self.GREENHOPPER_BASE_URL)
+        r = self._session.put(url, headers={'content-type': 'application/json'}, data=json.dumps(payload))
+        raise_on_error(r)
+
+        return json.loads(r.text)
+
     def completed_issues(self, board_id, sprint_id):
         """
         Return the completed issues for ``board_id`` and ``sprint_id``.
@@ -2136,22 +2220,40 @@ class GreenHopper(JIRA):
         raw_issue_json = json.loads(r.text)
         return Board(self._options, self._session, raw=raw_issue_json)
 
-    def create_sprint(self, name, board_id):
+    def create_sprint(self, name, board_id, startDate=None, endDate=None):
         """
         Create a new sprint for the ``board_id``.
 
         :param name: name of the sprint
         :param board_id: the board to add the sprint to
         """
+        url = self._get_url('sprint/%s' % board_id, base=self.GREENHOPPER_BASE_URL)
+        r = self._session.post(url, headers={'content-type': 'application/json'})
+        raise_on_error(r)
+        raw_issue_json = json.loads(r.text)
+        """ now r contains something like:
+        {
+              "id": 742,
+              "name": "Sprint 89",
+              "state": "FUTURE",
+              "linkedPagesCount": 0,
+              "startDate": "None",
+              "endDate": "None",
+              "completeDate": "None",
+              "remoteLinks": []
+        }"""
+
         payload = {}
         payload['name'] = name
-        payload['rapidViewId'] = board_id
-        payload['sprintMarkerId'] = 0
-        url = self._get_url('sprint/create', base=self.GREENHOPPER_BASE_URL)
-        r = self._session.post(url, headers={'content-type': 'application/json'}, data=json.dumps(payload))
+        if startDate:
+            payload["startDate"] = startDate
+        if endDate:
+            payload["endDate"] = endDate
+        url = self._get_url('sprint/%s' % raw_issue_json['id'], base=self.GREENHOPPER_BASE_URL)
+        r = self._session.put(url, headers={'content-type': 'application/json'}, data=json.dumps(payload))
         raise_on_error(r)
-
         raw_issue_json = json.loads(r.text)
+
         return Sprint(self._options, self._session, raw=raw_issue_json)
 
     def add_issues_to_sprint(self, sprint_id, issue_keys):
