@@ -1,8 +1,10 @@
+# -*- coding: utf-8 -*-
+from __future__ import unicode_literals
+from __future__ import print_function
 """
 This module implements the Resource classes that translate JSON from JIRA REST resources
 into usable objects.
 """
-from __future__ import print_function
 
 import re
 import sys
@@ -100,7 +102,7 @@ class Resource(object):
         headers = self._default_headers(headers)
         self._load(url, headers, params)
 
-    def update(self, async=False, jira=None, **kwargs):
+    def update(self, async=None, jira=None, **kwargs):
         """
         Update this resource on the server. Keyword arguments are marshalled into a dict before being sent. If this
         resource doesn't support ``PUT``, a :py:exc:`.JIRAError` will be raised; subclasses that specialize this method
@@ -108,6 +110,9 @@ class Resource(object):
 
         :param async: if true the request will be added to the queue so it can be executed later using async_run()
         """
+        if async is None:
+            async = self._options['async']
+
         data = {}
         for arg in kwargs:
             data[arg] = kwargs[arg]
@@ -157,15 +162,13 @@ class Resource(object):
                 #    data['fields']['assignee'] = {'name': self._options['autofix']}
             # EXPERIMENTAL --->
             # import grequests
-            # if async and 'grequests' in sys.modules:
-            #   if not hasattr(self._session, '_async_jobs'):
-            #        self._session._async_jobs = set()
-            #    self._session._async_jobs.add(grequests.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data)))
-            # else:
-            #    print "x", data
-            # <--- EXPERIMENTAL
-            r = self._session.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data))
-        raise_on_error(r)
+            if async and 'grequests' in sys.modules:
+                if not hasattr(self._session, '_async_jobs'):
+                    self._session._async_jobs = set()
+                self._session._async_jobs.add(grequests.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data)))
+            else:
+                r = self._session.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data))
+                raise_on_error(r)
         self._load(self.self)
 
     def delete(self, params=None):
@@ -174,8 +177,14 @@ class Resource(object):
         ``DELETE``, a :py:exc:`.JIRAError` will be raised; subclasses that specialize this method will only raise errors
         in case of user error.
         """
-        r = self._session.delete(self.self, params=params)
-        raise_on_error(r)
+
+        if self._options['async']:
+            if not hasattr(self._session, '_async_jobs'):
+                self._session._async_jobs = set()
+            self._session._async_jobs.add(grequests.delete(self.self, params=params))
+        else:
+            r = self._session.delete(self.self, params=params)
+            raise_on_error(r)
 
     def _load(self, url, headers=None, params=None):
         r = self._session.get(url, headers=headers, params=params)
@@ -277,7 +286,7 @@ class Issue(Resource):
         if raw:
             self._parse_raw(raw)
 
-    def update(self, fields=None, async=False, jira=None, **fieldargs):
+    def update(self, fields=None, async=None, jira=None, **fieldargs):
         """
         Update this issue on the server.
 
@@ -303,6 +312,15 @@ class Issue(Resource):
 
         super(Issue, self).update(async=async, jira=jira, **data)
 
+    def add_field_value(self, field, value):
+        """
+        Add a value to a field that supports multiple values, without resetting the existing values.
+
+        This should work with: labels, multiple checkbox lists, multiple select
+        """
+        field = self.instance.resolve_fields(field)
+        self.update({"update": {field: [{"add": value}]}})
+
     def delete(self, deleteSubtasks=False):
         """
         Delete this issue from the server.
@@ -629,7 +647,7 @@ def dict2resource(raw, top=None, options=None, session=None):
     or a ``PropertyHolder`` object (if no ``self`` link is present).
     """
     if top is None:
-        top = type('PropertyHolder', (object,), raw)
+        top = type(str('PropertyHolder'), (object,), raw)
 
     seqs = tuple, list, set, frozenset
     for i, j in iteritems(raw):
